@@ -1,26 +1,4 @@
-/*
-   MOD AUTH GSSAPI
-
-   Copyright (C) 2014 Simo Sorce <simo@redhat.com>
-
-   Permission is hereby granted, free of charge, to any person obtaining a
-   copy of this software and associated documentation files (the "Software"),
-   to deal in the Software without restriction, including without limitation
-   the rights to use, copy, modify, merge, publish, distribute, sublicense,
-   and/or sell copies of the Software, and to permit persons to whom the
-   Software is furnished to do so, subject to the following conditions:
-
-   The above copyright notice and this permission notice shall be included in
-   all copies or substantial portions of the Software.
-
-   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
-   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
-   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
-   DEALINGS IN THE SOFTWARE.
-*/
+/* Copyright (C) 2014, 2016 mod_auth_gssapi contributors - See COPYING for (C) terms */
 
 #include "mod_auth_gssapi.h"
 
@@ -224,10 +202,11 @@ static char *escape(apr_pool_t *pool, const char *name,
     return escaped;
 }
 
-char *mag_gss_name_to_ccache_name(request_rec *req,
-                                  char *dir, const char *gss_name)
+static char *get_ccache_name(request_rec *req, char *dir, const char *gss_name,
+                             bool use_unique, struct mag_conn *mc)
 {
-    char *escaped;
+    char *ccname, *escaped;
+    int ccachefd;
 
     /* We need to escape away '/', we can't have path separators in
      * a ccache file name */
@@ -236,22 +215,32 @@ char *mag_gss_name_to_ccache_name(request_rec *req,
     /* then escape away the separator (/) if any */
     escaped = escape(req->pool, escaped, '/', "~");
 
-    return apr_psprintf(req->pool, "%s/%s", dir, escaped);
+    if (use_unique == false) {
+        return apr_psprintf(req->pool, "%s/%s", dir, escaped);
+    }
+
+    ccname = apr_psprintf(mc->pool, "%s/%s-XXXXXX", dir, escaped);
+
+    ccachefd = mkstemp(ccname);
+    if (ccachefd == -1) {
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, req,
+                      "creating unique ccache file %s failed", ccname);
+        return NULL;
+    }
+    close(ccachefd);
+    return ccname;
 }
 
-static void mag_store_deleg_creds(request_rec *req,
-                                  char *dir, const char *gss_name,
+static void mag_store_deleg_creds(request_rec *req, const char *ccname,
                                   gss_cred_id_t delegated_cred)
 {
     gss_key_value_element_desc element;
     gss_key_value_set_desc store;
-    char *ccname;
     uint32_t maj, min;
     element.key = "ccache";
     store.elements = &element;
     store.count = 1;
 
-    ccname = mag_gss_name_to_ccache_name(req, dir, gss_name);
     element.value = apr_psprintf(req->pool, "FILE:%s", ccname);
 
     maj = gss_store_cred_into(&min, delegated_cred, GSS_C_INITIATE,
@@ -300,10 +289,12 @@ static bool is_mech_allowed(gss_OID_set allowed_mechs, gss_const_OID mech,
 #define AUTH_TYPE_NEGOTIATE 0
 #define AUTH_TYPE_BASIC 1
 #define AUTH_TYPE_RAW_NTLM 2
+#define AUTH_TYPE_IMPERSONATE 3
 const char *auth_types[] = {
     "Negotiate",
     "Basic",
     "NTLM",
+    "Impersonate",
     NULL
 };
 
@@ -358,6 +349,68 @@ gss_OID_set mag_filter_unwanted_mechs(gss_OID_set src)
     return src;
 }
 
+static uint32_t mag_context_loop(uint32_t *min,
+                                 request_rec *req,
+                                 gss_cred_id_t init_cred,
+                                 gss_cred_id_t accept_cred,
+                                 gss_OID mech_type,
+                                 uint32_t req_lifetime,
+                                 gss_name_t *client,
+                                 uint32_t *lifetime,
+                                 gss_cred_id_t *delegated_cred)
+{
+    gss_ctx_id_t init_ctx = GSS_C_NO_CONTEXT;
+    gss_ctx_id_t accept_ctx = GSS_C_NO_CONTEXT;
+    gss_buffer_desc init_token = GSS_C_EMPTY_BUFFER;
+    gss_buffer_desc accept_token = GSS_C_EMPTY_BUFFER;
+    gss_name_t accept_name = GSS_C_NO_NAME;
+    uint32_t maj, tmin;
+
+    maj = gss_inquire_cred_by_mech(min, accept_cred, mech_type, &accept_name,
+                                   NULL, NULL, NULL);
+    if (GSS_ERROR(maj)) {
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, req,
+                      "%s", mag_error(req, "gss_inquired_cred_by_mech() "
+                                      "failed", maj, *min));
+        return maj;
+    }
+
+    do {
+        /* output and input are inverted here, this is intentional */
+        maj = gss_init_sec_context(min, init_cred, &init_ctx,
+                                   accept_name, mech_type, GSS_C_DELEG_FLAG,
+                                   req_lifetime, GSS_C_NO_CHANNEL_BINDINGS,
+                                   &accept_token, NULL, &init_token, NULL,
+                                   NULL);
+        if (GSS_ERROR(maj)) {
+            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, req, "%s",
+                          mag_error(req, "gss_init_sec_context()", maj, *min));
+            goto done;
+        }
+        gss_release_buffer(&tmin, &accept_token);
+
+        maj = gss_accept_sec_context(min, &accept_ctx, accept_cred,
+                                     &init_token, GSS_C_NO_CHANNEL_BINDINGS,
+                                     client, NULL, &accept_token, NULL,
+                                     lifetime, delegated_cred);
+        if (GSS_ERROR(maj)) {
+            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, req, "%s",
+                          mag_error(req, "gss_accept_sec_context()",
+                                    maj, *min));
+            goto done;
+        }
+        gss_release_buffer(&tmin, &init_token);
+    } while (maj == GSS_S_CONTINUE_NEEDED);
+
+done:
+    gss_release_name(&tmin, &accept_name);
+    gss_release_buffer(&tmin, &init_token);
+    gss_release_buffer(&tmin, &accept_token);
+    gss_delete_sec_context(&tmin, &init_ctx, GSS_C_NO_BUFFER);
+    gss_delete_sec_context(&tmin, &accept_ctx, GSS_C_NO_BUFFER);
+    return maj;
+}
+
 static bool mag_auth_basic(request_rec *req,
                            struct mag_config *cfg,
                            gss_buffer_desc ba_user,
@@ -375,16 +428,10 @@ static bool mag_auth_basic(request_rec *req,
 #endif
     gss_name_t user = GSS_C_NO_NAME;
     gss_cred_id_t user_cred = GSS_C_NO_CREDENTIAL;
-    gss_ctx_id_t user_ctx = GSS_C_NO_CONTEXT;
-    gss_name_t server = GSS_C_NO_NAME;
     gss_cred_id_t server_cred = GSS_C_NO_CREDENTIAL;
-    gss_ctx_id_t server_ctx = GSS_C_NO_CONTEXT;
-    gss_buffer_desc input = GSS_C_EMPTY_BUFFER;
-    gss_buffer_desc output = GSS_C_EMPTY_BUFFER;
     gss_OID_set allowed_mechs;
     gss_OID_set filtered_mechs;
     gss_OID_set actual_mechs = GSS_C_NO_OID_SET;
-    uint32_t init_flags = 0;
     uint32_t maj, min;
     int present = 0;
     bool ret = false;
@@ -496,56 +543,10 @@ static bool mag_auth_basic(request_rec *req,
         goto done;
     }
 
-#ifdef HAVE_CRED_STORE
-    if (cfg->deleg_ccache_dir) {
-        /* delegate ourselves credentials so we store them as requested */
-        init_flags |= GSS_C_DELEG_FLAG;
-    }
-#endif
-
     for (int i = 0; i < actual_mechs->count; i++) {
-
-        /* free these if looping */
-        gss_release_buffer(&min, &output);
-        gss_release_buffer(&min, &input);
-        gss_release_name(&min, &server);
-
-        maj = gss_inquire_cred_by_mech(&min, server_cred,
-                                       &actual_mechs->elements[i],
-                                       &server, NULL, NULL, NULL);
-        if (GSS_ERROR(maj)) {
-            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, req,
-                          "%s", mag_error(req, "gss_inquired_cred_by_mech() "
-                                          "failed", maj, min));
-            continue;
-        }
-
-        do {
-            /* output and input are inverted here, this is intentional */
-            maj = gss_init_sec_context(&min, user_cred, &user_ctx, server,
-                                       &actual_mechs->elements[i], init_flags,
-                                       300, GSS_C_NO_CHANNEL_BINDINGS, &output,
-                                       NULL, &input, NULL, NULL);
-            if (GSS_ERROR(maj)) {
-                ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, req,
-                              "%s", mag_error(req, "gss_init_sec_context() "
-                                              "failed", maj, min));
-                break;
-            }
-            gss_release_buffer(&min, &output);
-            maj = gss_accept_sec_context(&min, &server_ctx, server_cred,
-                                         &input, GSS_C_NO_CHANNEL_BINDINGS,
-                                         client, mech_type, &output, NULL,
-                                         vtime, delegated_cred);
-            if (GSS_ERROR(maj)) {
-                ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, req,
-                              "%s", mag_error(req, "gss_accept_sec_context()"
-                                              " failed", maj, min));
-                break;
-            }
-            gss_release_buffer(&min, &input);
-        } while (maj == GSS_S_CONTINUE_NEEDED);
-
+        maj = mag_context_loop(&min, req, user_cred, server_cred,
+                               &actual_mechs->elements[i], 300, client, vtime,
+                               delegated_cred);
         if (maj == GSS_S_COMPLETE) {
             ret = true;
             break;
@@ -553,14 +554,9 @@ static bool mag_auth_basic(request_rec *req,
     }
 
 done:
-    gss_release_buffer(&min, &output);
-    gss_release_buffer(&min, &input);
-    gss_release_name(&min, &server);
-    gss_delete_sec_context(&min, &server_ctx, GSS_C_NO_BUFFER);
     gss_release_cred(&min, &server_cred);
     gss_release_name(&min, &user);
     gss_release_cred(&min, &user_cred);
-    gss_delete_sec_context(&min, &user_ctx, GSS_C_NO_BUFFER);
     gss_release_oid_set(&min, &actual_mechs);
     gss_release_oid_set(&min, &filtered_mechs);
 #ifdef HAVE_GSS_KRB5_CCACHE_NAME
@@ -629,8 +625,158 @@ static bool use_s4u2proxy(struct mag_req_cfg *req_cfg) {
     }
     return false;
 }
+
+static int mag_complete(struct mag_req_cfg *req_cfg, struct mag_conn *mc,
+                        gss_name_t client, gss_OID mech_type,
+                        uint32_t vtime, gss_cred_id_t delegated_cred);
+
+static apr_status_t mag_s4u2self(request_rec *req)
+{
+    apr_status_t ret = DECLINED;
+    const char *type;
+    struct mag_config *cfg;
+    struct mag_req_cfg *req_cfg;
+    gss_OID mech_type = discard_const(gss_mech_krb5);
+    gss_OID_set_desc gss_mech_krb5_set = { 1, mech_type };
+    gss_buffer_desc user_name = GSS_C_EMPTY_BUFFER;
+    gss_cred_id_t user_cred = GSS_C_NO_CREDENTIAL;
+    gss_name_t user = GSS_C_NO_NAME;
+    gss_name_t client = GSS_C_NO_NAME;
+    gss_cred_id_t server_cred = GSS_C_NO_CREDENTIAL;
+    gss_cred_id_t delegated_cred = GSS_C_NO_CREDENTIAL;
+    struct mag_conn *mc = NULL;
+    uint32_t vtime;
+    uint32_t maj, min;
+
+    req_cfg = mag_init_cfg(req);
+    cfg = req_cfg->cfg;
+
+    if (!cfg->s4u2self) {
+        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, req,
+                      "GSSapiImpersonate not On, skipping impersonation.");
+        return DECLINED;
+    }
+
+    type = ap_auth_type(req);
+    if (type && (strcasecmp(type, "GSSAPI") == 0)) {
+        /* do not try to impersonate if GSSAPI is handling real auth */
+        return DECLINED;
+    }
+
+    if (!req->user) {
+        ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, req,
+                      "Authentication user not found, "
+                      "skipping impersonation.");
+        return DECLINED;
+    }
+
+    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, req,
+                  "Using user %s for impersonation.", req->user);
+
+    if (!mag_acquire_creds(req, cfg, &gss_mech_krb5_set,
+                           GSS_C_BOTH, &server_cred, NULL)) {
+        goto done;
+    }
+
+    user_name.value = req->user;
+    user_name.length = strlen(user_name.value);
+    maj = gss_import_name(&min, &user_name, GSS_C_NT_USER_NAME, &user);
+    if (GSS_ERROR(maj)) {
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, req,
+                      "Failed to import user's name: %s",
+                      mag_error(req, "gss_import_name()",
+                                maj, min));
+        goto done;
+    }
+
+    maj = gss_acquire_cred_impersonate_name(&min, server_cred, user,
+                                            GSS_C_INDEFINITE,
+                                            &gss_mech_krb5_set,
+                                            GSS_C_INITIATE, &user_cred,
+                                            NULL, NULL);
+    if (GSS_ERROR(maj)) {
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, req,
+                      "Failed to impersonate %s: %s", req->user,
+                      mag_error(req, "gss_acquire_cred_impersonate_name()",
+                                maj, min));
+        goto done;
+    }
+
+    /* the following exchange is needed to decrypt the ticket and get named
+     * attributes as well as check if the ticket is forwardable when
+     * delegated credentials are requested */
+    maj = mag_context_loop(&min, req, user_cred, server_cred,
+                           discard_const(gss_mech_krb5), GSS_C_INDEFINITE,
+                           &client, &vtime, &delegated_cred);
+    if (GSS_ERROR(maj))
+        goto done;
+
+    if (cfg->deleg_ccache_dir && delegated_cred == GSS_C_NO_CREDENTIAL) {
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, req,
+                      "Failed to obtain delegated credentials, "
+                      "does service have +ok_to_auth_as_delegate?");
+        goto done;
+    }
+
+    mc = mag_new_conn_ctx(req->pool);
+    mc->auth_type = AUTH_TYPE_IMPERSONATE;
+
+    ret = mag_complete(req_cfg, mc, client, mech_type, vtime, delegated_cred);
+    if (ret != OK) ret = DECLINED;
+
+done:
+    gss_release_cred(&min, &user_cred);
+    gss_release_name(&min, &user);
+    gss_release_name(&min, &client);
+    gss_release_cred(&min, &server_cred);
+    gss_release_cred(&min, &delegated_cred);
+    return ret;
+}
 #endif
 
+static apr_status_t mag_oid_set_destroy(void *ptr)
+{
+    uint32_t min;
+    gss_OID_set set = (gss_OID_set)ptr;
+    (void)gss_release_oid_set(&min, &set);
+    return APR_SUCCESS;
+}
+
+static gss_OID_set mag_get_negotiate_mechs(apr_pool_t *p, gss_OID_set desired)
+{
+    gss_OID spnego_oid = discard_const(&gss_mech_spnego);
+    uint32_t maj, min;
+    int present = 0;
+
+    maj = gss_test_oid_set_member(&min, spnego_oid, desired, &present);
+    if (maj != GSS_S_COMPLETE) {
+        return GSS_C_NO_OID_SET;
+    }
+    if (present) {
+        return desired;
+    } else {
+        gss_OID_set set;
+        maj = gss_create_empty_oid_set(&min, &set);
+        if (maj != GSS_S_COMPLETE) {
+            return GSS_C_NO_OID_SET;
+        }
+        apr_pool_cleanup_register(p, (void *)set,
+                                  mag_oid_set_destroy,
+                                  apr_pool_cleanup_null);
+        maj = gss_add_oid_set_member(&min, spnego_oid, &set);
+        if (maj != GSS_S_COMPLETE) {
+            return GSS_C_NO_OID_SET;
+        }
+        for (int i = 0; i < desired->count; i++) {
+             maj = gss_add_oid_set_member(&min, &desired->elements[i], &set);
+            if (maj != GSS_S_COMPLETE) {
+                return GSS_C_NO_OID_SET;
+            }
+        }
+        return set;
+    }
+}
+
 static int mag_auth(request_rec *req)
 {
     const char *type;
@@ -644,7 +790,6 @@ static int mag_auth(request_rec *req)
     gss_ctx_id_t *pctx;
     gss_buffer_desc input = GSS_C_EMPTY_BUFFER;
     gss_buffer_desc output = GSS_C_EMPTY_BUFFER;
-    gss_buffer_desc name = GSS_C_EMPTY_BUFFER;
     gss_buffer_desc ba_user;
     gss_buffer_desc ba_pwd;
     gss_name_t client = GSS_C_NO_NAME;
@@ -657,7 +802,6 @@ static int mag_auth(request_rec *req)
     size_t replen;
     gss_OID mech_type = GSS_C_NO_OID;
     gss_OID_set desired_mechs = GSS_C_NO_OID_SET;
-    gss_buffer_desc lname = GSS_C_EMPTY_BUFFER;
     struct mag_conn *mc = NULL;
     int i;
     bool send_auth_header = true;
@@ -671,7 +815,13 @@ static int mag_auth(request_rec *req)
 
     cfg = req_cfg->cfg;
 
-    desired_mechs = req_cfg->desired_mechs;
+    if ((req_cfg->desired_mechs == GSS_C_NO_OID_SET) ||
+        (req_cfg->desired_mechs->count == 0)) {
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, req,
+                      "List of desired mechs is missing or empty, "
+                      "can't proceed!");
+        return HTTP_UNAUTHORIZED;
+    }
 
     /* implicit auth for subrequests if main auth already happened */
     if (!ap_is_initial_req(req) && req->main != NULL) {
@@ -767,6 +917,13 @@ static int mag_auth(request_rec *req)
         if (!parse_auth_header(req->pool, &auth_header, &input)) {
             goto done;
         }
+        desired_mechs = mag_get_negotiate_mechs(req->pool,
+                                                req_cfg->desired_mechs);
+        if (desired_mechs == GSS_C_NO_OID_SET) {
+            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, req,
+                          "Failed to get negotiate_mechs");
+            goto done;
+        }
         break;
     case AUTH_TYPE_BASIC:
         if (!cfg->use_basic_auth) {
@@ -812,6 +969,11 @@ static int mag_auth(request_rec *req)
         }
 
         desired_mechs = discard_const(gss_mech_set_ntlmssp);
+        if (desired_mechs == GSS_C_NO_OID_SET) {
+            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, req,
+                          "No support for ntlmssp mech");
+            goto done;
+        }
         break;
 
     default:
@@ -837,7 +999,11 @@ static int mag_auth(request_rec *req)
         if (mag_auth_basic(req, cfg, ba_user, ba_pwd,
                            &client, &mech_type,
                            &delegated_cred, &vtime)) {
-            goto complete;
+
+            ret = mag_complete(req_cfg, mc, client, mech_type, vtime,
+                               delegated_cred);
+            if (ret == OK)
+                mag_basic_cache(req_cfg, mc, ba_user, ba_pwd);
         }
         goto done;
     }
@@ -880,62 +1046,13 @@ static int mag_auth(request_rec *req)
         goto done;
     }
 
-complete:
-    maj = gss_display_name(&min, client, &name, NULL);
-    if (GSS_ERROR(maj)) {
-        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, req, "%s",
-                      mag_error(req, "gss_display_name() failed",
-                                maj, min));
-        goto done;
-    }
+    ret = mag_complete(req_cfg, mc, client, mech_type, vtime, delegated_cred);
 
-    mc->gss_name = apr_pstrndup(req->pool, name.value, name.length);
-    if (vtime == GSS_C_INDEFINITE || vtime < MIN_SESS_EXP_TIME) {
-        vtime = MIN_SESS_EXP_TIME;
-    }
-    mc->expiration = time(NULL) + vtime;
-
-    mag_get_name_attributes(req, cfg, client, mc);
-
-#ifdef HAVE_CRED_STORE
-    if (cfg->deleg_ccache_dir && delegated_cred != GSS_C_NO_CREDENTIAL) {
-        mag_store_deleg_creds(req, cfg->deleg_ccache_dir, mc->gss_name,
-                              delegated_cred);
-        mc->delegated = true;
-    }
-#endif
-
-    if (cfg->map_to_local) {
-        maj = gss_localname(&min, client, mech_type, &lname);
-        if (maj != GSS_S_COMPLETE) {
-            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, req, "%s",
-                          mag_error(req, "gss_localname() failed", maj, min));
-            goto done;
-        }
-        mc->user_name = apr_pstrndup(req->pool, lname.value, lname.length);
-    } else {
-        mc->user_name = apr_pstrdup(mc->pool, mc->gss_name);
-    }
-
-    mc->established = true;
-    if (auth_type == AUTH_TYPE_BASIC) {
-        mag_basic_cache(req_cfg, mc, ba_user, ba_pwd);
-    }
-    if (req_cfg->use_sessions) {
-        mag_attempt_session(req_cfg, mc);
-    }
-
-    /* Now set request data and env vars */
-    mag_set_req_data(req, cfg, mc);
-
-    if (req_cfg->send_persist)
+    if (ret == OK && req_cfg->send_persist)
         apr_table_set(req->headers_out, "Persistent-Auth",
             cfg->gss_conn_ctx ? "true" : "false");
 
-    ret = OK;
-
 done:
-
     if ((auth_type != AUTH_TYPE_BASIC) && (output.length != 0)) {
         int prefixlen = strlen(mag_str_auth_type(auth_type)) + 1;
         replen = apr_base64_encode_len(output.length) + 1;
@@ -969,12 +1086,94 @@ done:
     gss_release_cred(&min, &delegated_cred);
     gss_release_buffer(&min, &output);
     gss_release_name(&min, &client);
+    return ret;
+}
+
+static int mag_complete(struct mag_req_cfg *req_cfg, struct mag_conn *mc,
+                        gss_name_t client, gss_OID mech_type,
+                        uint32_t vtime, gss_cred_id_t delegated_cred)
+{
+    gss_buffer_desc lname = GSS_C_EMPTY_BUFFER;
+    gss_buffer_desc name = GSS_C_EMPTY_BUFFER;
+    struct mag_config *cfg = req_cfg->cfg;
+    request_rec *req = req_cfg->req;
+    int ret = HTTP_UNAUTHORIZED;
+    uint32_t maj, min;
+
+    maj = gss_display_name(&min, client, &name, NULL);
+    if (GSS_ERROR(maj)) {
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, req, "%s",
+                      mag_error(req, "gss_display_name() failed",
+                                maj, min));
+        goto done;
+    }
+
+    mc->gss_name = apr_pstrndup(req->pool, name.value, name.length);
+    if (vtime == GSS_C_INDEFINITE || vtime < MIN_SESS_EXP_TIME) {
+        vtime = MIN_SESS_EXP_TIME;
+    }
+    mc->expiration = time(NULL) + vtime;
+
+    mag_get_name_attributes(req, cfg, client, mc);
+
+#ifdef HAVE_CRED_STORE
+    if (cfg->deleg_ccache_dir &&
+        delegated_cred != GSS_C_NO_CREDENTIAL) {
+        char *ccache_path;
+
+        mc->ccname = 0;
+        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, req,
+                      "requester: %s", mc->gss_name);
+
+        ccache_path = get_ccache_name(req, cfg->deleg_ccache_dir, mc->gss_name,
+                                      cfg->deleg_ccache_unique, mc);
+        if (ccache_path == NULL) {
+            goto done;
+        }
+
+        mag_store_deleg_creds(req, ccache_path, delegated_cred);
+        mc->delegated = true;
+
+        if (!req_cfg->use_sessions && cfg->deleg_ccache_unique) {
+            /* queue removing ccache to avoid littering filesystem */
+            apr_pool_cleanup_register(mc->pool, ccache_path,
+                                      (int (*)(void *)) unlink,
+                                      apr_pool_cleanup_null);
+        }
+
+        /* extract filename from full path */
+        mc->ccname = strrchr(ccache_path, '/') + 1;
+    }
+#endif
+
+    if (cfg->map_to_local) {
+        maj = gss_localname(&min, client, mech_type, &lname);
+        if (maj != GSS_S_COMPLETE) {
+            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, req, "%s",
+                          mag_error(req, "gss_localname() failed", maj, min));
+            goto done;
+        }
+        mc->user_name = apr_pstrndup(mc->pool, lname.value, lname.length);
+    } else {
+        mc->user_name = apr_pstrdup(mc->pool, mc->gss_name);
+    }
+
+    mc->established = true;
+    if (req_cfg->use_sessions) {
+        mag_attempt_session(req_cfg, mc);
+    }
+
+    /* Now set request data and env vars */
+    mag_set_req_data(req, cfg, mc);
+
+    ret = OK;
+
+done:
     gss_release_buffer(&min, &name);
     gss_release_buffer(&min, &lname);
     return ret;
 }
 
-
 static void *mag_create_dir_config(apr_pool_t *p, char *dir)
 {
     struct mag_config *cfg;
@@ -1028,6 +1227,15 @@ static const char *mag_use_s4u2p(cmd_parms *parms, void *mconfig, int on)
 
     return NULL;
 }
+
+static const char *mag_deleg_ccache_unique(cmd_parms *parms, void *mconfig,
+                                           int on)
+{
+    struct mag_config *cfg = (struct mag_config *)mconfig;
+    cfg->deleg_ccache_unique = on ? true : false;
+    return NULL;
+}
+
 #endif
 
 static const char *mag_sess_key(cmd_parms *parms, void *mconfig, const char *w)
@@ -1136,16 +1344,8 @@ static const char *mag_use_basic_auth(cmd_parms *parms, void *mconfig, int on)
 }
 #endif
 
-static apr_status_t mag_oid_set_destroy(void *ptr)
-{
-    uint32_t min;
-    gss_OID_set set = (gss_OID_set)ptr;
-    (void)gss_release_oid_set(&min, &set);
-    return APR_SUCCESS;
-}
-
 static bool mag_list_of_mechs(cmd_parms *parms, gss_OID_set *oidset,
-                              bool add_spnego, const char *w)
+                              const char *w)
 {
     gss_buffer_desc buf = { 0 };
     uint32_t maj, min;
@@ -1161,17 +1361,6 @@ static bool mag_list_of_mechs(cmd_parms *parms, gss_OID_set *oidset,
             *oidset = GSS_C_NO_OID_SET;
             return false;
         }
-        if (add_spnego) {
-            oid = discard_const(&gss_mech_spnego);
-            maj = gss_add_oid_set_member(&min, oid, &set);
-            if (maj != GSS_S_COMPLETE) {
-                ap_log_error(APLOG_MARK, APLOG_ERR, 0, parms->server,
-                             "gss_add_oid_set_member() failed.");
-                (void)gss_release_oid_set(&min, &set);
-                *oidset = GSS_C_NO_OID_SET;
-                return false;
-            }
-        }
         /* register in the pool so it can be released once the server
          * winds down */
         apr_pool_cleanup_register(parms->pool, (void *)set,
@@ -1216,7 +1405,7 @@ static const char *mag_allow_mech(cmd_parms *parms, void *mconfig,
 {
     struct mag_config *cfg = (struct mag_config *)mconfig;
 
-    if (!mag_list_of_mechs(parms, &cfg->allowed_mechs, true, w))
+    if (!mag_list_of_mechs(parms, &cfg->allowed_mechs, w))
         return "Failed to apply GssapiAllowedMech directive";
 
     return NULL;
@@ -1298,7 +1487,7 @@ static const char *mag_basic_auth_mechs(cmd_parms *parms, void *mconfig,
 {
     struct mag_config *cfg = (struct mag_config *)mconfig;
 
-    if (!mag_list_of_mechs(parms, &cfg->basic_mechs, false, w))
+    if (!mag_list_of_mechs(parms, &cfg->basic_mechs, w))
         return "Failed to apply GssapiBasicAuthMech directive";
 
     return NULL;
@@ -1352,6 +1541,12 @@ static const command_rec mag_commands[] = {
                     "Credential Store"),
     AP_INIT_RAW_ARGS("GssapiDelegCcacheDir", mag_deleg_ccache_dir, NULL,
                      OR_AUTHCFG, "Directory to store delegated credentials"),
+    AP_INIT_FLAG("GssapiDelegCcacheUnique", mag_deleg_ccache_unique, NULL,
+                 OR_AUTHCFG, "Use unique ccaches for delgation"),
+    AP_INIT_FLAG("GssapiImpersonate", ap_set_flag_slot,
+          (void *)APR_OFFSETOF(struct mag_config, s4u2self), OR_AUTHCFG,
+               "Do impersonation call (S4U2Self) "
+               "based on already authentication username"),
 #endif
 #ifdef HAVE_GSS_ACQUIRE_CRED_WITH_PASSWORD
     AP_INIT_FLAG("GssapiBasicAuth", mag_use_basic_auth, NULL, OR_AUTHCFG,
@@ -1371,9 +1566,17 @@ static const command_rec mag_commands[] = {
 static void
 mag_register_hooks(apr_pool_t *p)
 {
+#ifdef AP_AUTH_INTERNAL_PER_CONF
+    ap_hook_check_authn(mag_auth, NULL, NULL, APR_HOOK_MIDDLE,
+                                                AP_AUTH_INTERNAL_PER_CONF);
+#else
     ap_hook_check_user_id(mag_auth, NULL, NULL, APR_HOOK_MIDDLE);
+#endif
     ap_hook_post_config(mag_post_config, NULL, NULL, APR_HOOK_MIDDLE);
     ap_hook_pre_connection(mag_pre_connection, NULL, NULL, APR_HOOK_MIDDLE);
+#ifdef HAVE_CRED_STORE
+    ap_hook_fixups(mag_s4u2self, NULL, NULL, APR_HOOK_MIDDLE);
+#endif
 }
 
 module AP_MODULE_DECLARE_DATA auth_gssapi_module =
